MODULE MESSY_MAIN_GRID_TOOLS

! ------------------------------------------------------------------
! Author: Patrick Joeckel, MPICH, Mainz, June 2002
!         Astrid  Kerkweg, UniMz, Mainz, 2012-2013
!         -> re-structured for more general GRID application
! ******************************************************************

  USE MESSY_MAIN_CONSTANTS_MEM, ONLY: dp, r_e=> radius_earth

  IMPLICIT NONE

  PUBLIC  :: RGTOOL_CONVERT         ! CONVERTS NCVAR INTO 4D-ARRAY
  PUBLIC  :: RGTOOL_CONVERT_DAT2VAR ! CONVERTS 4D-ARRAY INTO NCVAR
  PUBLIC  :: RGTOOL_G2C             ! CONVERTS GRID-INFORMATION INTO ARRAYS
  ! um_ak_20150506+
  PUBLIC :: ALINE_ARRAY
  PUBLIC :: DEALINE_ARRAY
  ! um_ak_20150506-
  PUBLIC  :: SET_SURFACE_PRESSURE   ! SET CURRENT SURFACE PRESSURE FOR VERTICAL
                                    ! GRID um_ak_20141007

CONTAINS

! ----------------------------------------------------------------------
! ******************************************************************
! ----------------------------------------------------------------------
  SUBROUTINE RGTOOL_CONVERT(var, dat, grid, order)

    ! CONVERTS NCREGRID OUTPUT OF TYPE N-ARRAY (var) ON GRID
    ! grid TO 4D-ARRAY (dat)
    ! THE OPTIONAL STRING order DEFINES THE ORDER OF DIMENSIONS
    ! (DEFAULT: 'xyzn')
    !
    ! Author: Patrick Joeckel, MPIC, October 2002

    USE MESSY_MAIN_GRID_NETCDF,  ONLY: t_ncvar, RGMSG , RGMLE, RGMLI, RGMLIC &
                                     , ERRMSG, QCMP_NCDIM, VTYPE_DOUBLE    &
                                     , VTYPE_INT, VTYPE_BYTE, VTYPE_CHAR   &
                                     , VTYPE_UNDEF, VTYPE_REAL, ELEMENT
    USE MESSY_MAIN_GRID,         ONLY: t_geohybgrid
  
    ! REGRID MODULES

    IMPLICIT NONE

    INTRINSIC :: ASSOCIATED, INDEX, PRESENT, PRODUCT, REAL, SIZE, SUM, TRIM

    ! I/O
    TYPE (t_ncvar),    INTENT(IN)           :: var   ! nc-variable
    REAL(dp), DIMENSION(:,:,:,:), POINTER   :: dat   ! DATA ON DESTINATION GRID
    ! um_ak_20150506+
    !TYPE(t_geohybgrid),INTENT(IN), OPTIONAL :: grid  ! grid information
    TYPE(t_geohybgrid),INTENT(IN)           :: grid  ! grid information
    ! um_ak_20150506-
    CHARACTER(LEN=4),  INTENT(IN), OPTIONAL :: order ! DEFAULT: 'xyzn'

    ! LOCAL
    CHARACTER(LEN=*), PARAMETER :: substr = 'RGTOOL_CONVERT'
    INTEGER :: ovec(4)          ! order of (x,y,z,n) for output field
    INTEGER :: ldvar(4)         ! ordered dim-lenghts in var
    INTEGER :: dpos(3)          ! position of lon, lat, lev in variable
    INTEGER :: nrvdim           ! number of recognized dimensions in variable
    INTEGER :: npdlv            ! product of dimension-lengths in variable
    CHARACTER(LEN=6) :: ostr(4) ! order-string for output
    INTEGER :: i                ! counter
    INTEGER :: nvec(4)          ! counter parameter for dimensions
    INTEGER :: vtype            ! type of variable
    INTEGER, DIMENSION(:), POINTER :: vec   ! element vector
    INTEGER :: status           ! status flag

    ! INIT
    NULLIFY(vec)
    ovec = (/ 1,2,3,4 /)                               ! DEFAULT
    ostr = (/ 'x/LON ', 'y/LAT ', 'z/LEV ', 'n/PAR '/) ! DEFAULT
    IF (PRESENT(order)) THEN
       ovec(1) = INDEX(order, 'x')
       ostr(ovec(1)) = 'x/LON '
       ovec(2) = INDEX(order, 'y')
       ostr(ovec(2)) = 'y/LAT '
       ovec(3) = INDEX(order, 'z')
       ostr(ovec(3)) = 'z/LEV '
       ovec(4) = INDEX(order, 'n')
       ostr(ovec(4)) = 'n/PAR '
       IF (SUM(ovec) /= 10) THEN
          CALL RGMSG(substr, RGMLE, &
               'ERROR IN '''//TRIM(order)//'''-STRING AT SUBROUTINE CALL !')
       END IF
    END IF
    !
    IF (ASSOCIATED(dat)) THEN
       DEALLOCATE(dat, STAT=status)
       CALL ERRMSG(substr,status,1)
    END IF
    NULLIFY(dat)
    !
    dpos(:) = 0
    nrvdim = 0
    ldvar(:) = 1
    IF (ASSOCIATED(var%dat%dim)) THEN
       npdlv = PRODUCT(var%dat%dim)
    ELSE
       RETURN
    END IF

    ! LOOP OVER VARIABLE DIMENSIONS AND CHECK NAMES, LENGTHS, AND POSITIONS
    DO i=1, var%ndims
       IF (ASSOCIATED(grid%lonm%dim)) THEN
          IF (QCMP_NCDIM(var%dim(i), grid%lonm%dim(1)) > 1) THEN
             nrvdim = nrvdim+1
             ldvar(ovec(1)) = var%dim(i)%len
             dpos(1) = i
          END IF
       END IF
       IF (ASSOCIATED(grid%latm%dim)) THEN
          IF (QCMP_NCDIM(var%dim(i), grid%latm%dim(1)) > 1) THEN
             nrvdim = nrvdim+1
             ldvar(ovec(2)) = var%dim(i)%len
             dpos(2) = i
          END IF
       END IF
       IF (ASSOCIATED(grid%clonm%dim)) THEN
          IF (QCMP_NCDIM(var%dim(i), grid%clonm%dim(1)) > 1) THEN
             nrvdim = nrvdim+1
             ldvar(ovec(1)) = var%dim(i)%len
             dpos(1) = i
          END IF
       END IF
       IF (ASSOCIATED(grid%clatm%dim)) THEN
          IF (QCMP_NCDIM(var%dim(i), grid%clatm%dim(2)) > 1) THEN
             nrvdim = nrvdim+1
             ldvar(ovec(2)) = var%dim(i)%len
             dpos(2) = i
          END IF
       END IF
! um_ak_20150605+
       IF (ASSOCIATED(grid%rlonm%dim)) THEN
          IF (QCMP_NCDIM(var%dim(i), grid%rlonm%dim(1)) > 1) THEN
             nrvdim = nrvdim+1
             ldvar(ovec(1)) = var%dim(i)%len
             dpos(1) = i
          END IF
       END IF
       IF (ASSOCIATED(grid%rlatm%dim)) THEN
          IF (QCMP_NCDIM(var%dim(i), grid%rlatm%dim(1)) > 1) THEN
             nrvdim = nrvdim+1
             ldvar(ovec(2)) = var%dim(i)%len
             dpos(2) = i
          END IF
       END IF
! um_ak_20150605-
       IF (ASSOCIATED(grid%hyam%dim)) THEN
          IF (QCMP_NCDIM(var%dim(i), grid%hyam%dim(1)) > 1) THEN
             nrvdim = nrvdim+1
             ldvar(ovec(3)) = var%dim(i)%len
             dpos(3) = i
          END IF
       ELSE
          IF (ASSOCIATED(grid%hybm%dim)) THEN
             IF (QCMP_NCDIM(var%dim(i), grid%hybm%dim(1)) > 1) THEN
                nrvdim = nrvdim+1
                ldvar(ovec(3)) = var%dim(i)%len
                dpos(3) = i
             END IF
          END IF
       END IF
    END DO

    ldvar(ovec(4)) = npdlv/(ldvar(ovec(1))*ldvar(ovec(2))*ldvar(ovec(3)))

    CALL RGMSG(substr, RGMLI, 'VARIABLE '''//TRIM(var%name)//''' IS ' &
         ,var%ndims,'-DIMENSIONAL')
    CALL RGMSG(substr, RGMLIC, '(',var%dat%dim,')')
    CALL RGMSG(substr, RGMLIC, 'DATA DIMENSION-LENGHTS ARE: ')
    DO i=1, SIZE(ostr)
       CALL RGMSG(substr, RGMLIC, '  '//ostr(i)//': ',ldvar(i),' ')
    END DO
    CALL RGMSG(substr, RGMLIC, &
         'SOURCE POSITIONS (x/LON, y/LAT, z/LEV) ARE: ',dpos,' ')

    ! ALLOCATE DATA SPACE
    ALLOCATE(dat(ldvar(1),ldvar(2),ldvar(3),ldvar(4)), STAT=status)
    CALL ERRMSG(substr,status,1)

    ! COPY DATA (LOOP OVER ELEMENTS)
    vtype = var%dat%type
    DO i=1, npdlv
       CALL ELEMENT(var%dat%dim, i, vec)
       nvec((ovec(4))) = PRODUCT(vec)
       IF (dpos(1) /= 0) THEN
          nvec((ovec(1))) = vec(dpos(1))
          nvec(ovec(4)) = nvec(ovec(4))/nvec(ovec(1))
       ELSE
          nvec(ovec(1)) = 1
       END IF
       IF (dpos(2) /= 0) THEN
          nvec((ovec(2))) = vec(dpos(2))
          nvec(ovec(4)) = nvec(ovec(4))/nvec(ovec(2))
       ELSE
          nvec(ovec(2)) = 1
       END IF
       IF (dpos(3) /= 0) THEN
          nvec((ovec(3))) = vec(dpos(3))
          nvec(ovec(4)) = nvec(ovec(4))/nvec(ovec(3))
       ELSE
          nvec(ovec(3)) = 1
       END IF
       SELECT CASE(vtype)
       CASE(VTYPE_DOUBLE)
          dat(nvec(1), nvec(2), nvec(3), nvec(4)) &
               = var%dat%vd(i)
       CASE(VTYPE_REAL)
          dat(nvec(1), nvec(2), nvec(3), nvec(4)) &
               = REAL(var%dat%vr(i),dp)
       CASE(VTYPE_INT)
          dat(nvec(1), nvec(2), nvec(3), nvec(4)) &
               = REAL(var%dat%vi(i),dp)
       CASE(VTYPE_BYTE)
          dat(nvec(1), nvec(2), nvec(3), nvec(4)) &
               = REAL(var%dat%vb(i),dp)
       CASE(VTYPE_CHAR)
          CALL RGMSG(substr, RGMLE, 'DATA OF TYPE CHAR NOT SUPPORTED !')
       CASE(VTYPE_UNDEF)
          CALL RGMSG(substr, RGMLE, 'UNDEFINED TYPE OF DATA !')
       CASE DEFAULT
          CALL RGMSG(substr, RGMLE, 'UNRECOGNIZED TYPE OF DATA !')
       END SELECT
       !
       DEALLOCATE(vec, STAT=status)
       NULLIFY(vec) ! op_pj_20121121
       CALL ERRMSG(substr,status,2)
    END DO

  END SUBROUTINE RGTOOL_CONVERT
! ----------------------------------------------------------------------

! ----------------------------------------------------------------------
  SUBROUTINE RGTOOL_CONVERT_DAT2VAR(var, dat, vname, grid, order)

    ! CONVERTS 4D-ARRAY (dat) TO TYPE N-ARRAY (var) ON GRID
    ! THE OPTIONAL STRING order DEFINES THE ORDER OF DIMENSIONS
    ! (DEFAULT: 'xyzn')
    !
    ! Author: Andreas Baumgaertner, MPIC, July 2010

    USE MESSY_MAIN_GRID_NETCDF,  ONLY: t_ncvar, RGMSG , RGMLE              &
                                     , ERRMSG, INIT_NCVAR, RENAME_NCVAR    &
                                     , QDEF_NCVAR, COPY_NCDIM, INIT_NARRAY &
                                     , ELEMENT, NF90_DOUBLE, VTYPE_DOUBLE
    USE MESSY_MAIN_GRID,         ONLY: t_geohybgrid

    IMPLICIT NONE

    INTRINSIC :: ASSOCIATED, INDEX, PRESENT, PRODUCT, SIZE, SUM, TRIM

    ! I/O
    ! um_ak_20140312+
    !TYPE (t_ncvar),     INTENT(OUT)         :: var   ! nc-variable 
    TYPE (t_ncvar),      INTENT(INOUT)       :: var   ! nc-variable 
    ! um_ak_20140312-
    REAL(DP), DIMENSION(:,:,:,:), POINTER    :: dat   ! DATA ON DESTINATION GRID
    CHARACTER(LEN=*),    INTENT(IN)          :: vname ! name of variable
    TYPE (t_geohybgrid), INTENT(IN)          :: grid  ! grid information
    CHARACTER(LEN=4),    INTENT(IN),OPTIONAL :: order ! DEFAULT: 'xyzn'

    ! LOCAL
    CHARACTER(LEN=*), PARAMETER :: substr = 'RGTOOL_CONVERT_DAT2VAR'
    INTEGER :: ovec(4)          ! order of (x,y,z,n) for input field
    INTEGER :: ldvar(4)         ! ordered dim-lenghts in var
    INTEGER :: dpos(3)          ! position of lon, lat, lev in variable
    INTEGER :: nrvdim           ! number of recognized dimensions in variable
    INTEGER :: npdlv            ! product of dimension-lengths in variable
    !CHARACTER(LEN=6) :: ostr(4)! order-string for output
    INTEGER :: i,j              ! counter
    INTEGER :: nvec(4)          ! counter parameter for dimensions
    INTEGER, DIMENSION(:), POINTER     :: vec     ! element vector
    INTEGER, DIMENSION(:), ALLOCATABLE :: zdimlen ! local dim. lengths
    INTEGER :: status           ! status flag

    ! INIT
    CALL INIT_NCVAR(var)
    CALL RENAME_NCVAR(var, vname)
    var%xtype = NF90_DOUBLE ! xtype ?
    ! CREATE VARIABLE DIMENSIONS FROM GRID 
    nrvdim = 0
    IF (QDEF_NCVAR(grid%lonm)) THEN
       nrvdim = nrvdim + 1
    ELSE IF (QDEF_NCVAR(grid%clonm)) THEN
       nrvdim = nrvdim + 1
    ENDIF
    IF (QDEF_NCVAR(grid%latm)) THEN
       nrvdim = nrvdim + 1
    ELSE IF (QDEF_NCVAR(grid%clatm)) THEN
       nrvdim = nrvdim + 1
    ENDIF
    IF (QDEF_NCVAR(grid%hyam) .OR. QDEF_NCVAR(grid%hybm)) THEN
       nrvdim = nrvdim + 1
    ELSE IF (QDEF_NCVAR(grid%hyai) .OR. QDEF_NCVAR(grid%hybi)) THEN
       nrvdim = nrvdim + 1
    ENDIF
    IF (QDEF_NCVAR(grid%timem)) nrvdim = nrvdim + 1

    var%ndims = nrvdim
    ALLOCATE(var%dim(var%ndims))
    
    NULLIFY(vec)
    ovec = (/ 1,2,3,4 /)                                ! DEFAULT
    !ostr = (/ 'x/LON ', 'y/LAT ', 'z/LEV ', 'n/PAR '/) ! DEFAULT
    IF (PRESENT(order)) THEN
       ovec(1) = INDEX(order, 'x')
       !ostr(ovec(1)) = 'x/LON '
       ovec(2) = INDEX(order, 'y')
       !ostr(ovec(2)) = 'y/LAT '
       ovec(3) = INDEX(order, 'z')
       !ostr(ovec(3)) = 'z/LEV '
       ovec(4) = INDEX(order, 'n')
       !ostr(ovec(4)) = 'n/PAR '
       IF (SUM(ovec) /= 10) THEN
          CALL RGMSG(substr, RGMLE, &
               'ERROR IN '''//TRIM(order)//'''-STRING AT SUBROUTINE CALL !')
       END IF
    END IF
    !
    !
    dpos(:) = 0
    nrvdim = 0
 
    ! LOOP OVER VARIABLE DIMENSIONS AND CHECK NAMES, LENGTHS, AND POSITIONS
    IF (ASSOCIATED(grid%lonm%dim)) THEN
       nrvdim = nrvdim + 1
       CALL COPY_NCDIM(var%dim(nrvdim), grid%lonm%dim(1))
       ldvar(ovec(1)) = var%dim(nrvdim)%len
       dpos(1) = 1 ! BUT THIS SHOULD BE DETERMINED BY ORDER
    ELSE IF (ASSOCIATED(grid%clonm%dim)) THEN
       nrvdim = nrvdim + 1
       CALL COPY_NCDIM(var%dim(nrvdim), grid%clonm%dim(1))
       ldvar(ovec(1)) = var%dim(nrvdim)%len
       dpos(1) = 1 ! BUT THIS SHOULD BE DETERMINED BY ORDER
    END IF
    IF (ASSOCIATED(grid%latm%dim)) THEN
       nrvdim = nrvdim + 1
       CALL COPY_NCDIM(var%dim(nrvdim), grid%latm%dim(1))
       ldvar(ovec(2)) = var%dim(nrvdim)%len
       dpos(2) = 2
    ELSE IF (ASSOCIATED(grid%clatm%dim)) THEN
       nrvdim = nrvdim + 1
       CALL COPY_NCDIM(var%dim(nrvdim), grid%clatm%dim(2))
       ldvar(ovec(2)) = var%dim(nrvdim)%len
       dpos(2) = 2
    END IF
    IF (ASSOCIATED(grid%hyam%dim)) THEN
       nrvdim = nrvdim + 1
       CALL COPY_NCDIM(var%dim(nrvdim), grid%hyam%dim(1))
       ldvar(ovec(3)) = var%dim(nrvdim)%len
       dpos(3) = 3
    ELSE IF (ASSOCIATED(grid%hybm%dim)) THEN
       nrvdim = nrvdim + 1
       CALL COPY_NCDIM(var%dim(nrvdim), grid%hybm%dim(1))
       ldvar(ovec(3)) = var%dim(nrvdim)%len
       dpos(3) = 3
    ELSE IF (ASSOCIATED(grid%hyai%dim)) THEN
       nrvdim = nrvdim + 1
       CALL COPY_NCDIM(var%dim(nrvdim), grid%hyai%dim(1))
       ldvar(ovec(3)) = var%dim(nrvdim)%len
       dpos(3) = 3
    ELSE IF (ASSOCIATED(grid%hybi%dim)) THEN
       nrvdim = nrvdim + 1
       CALL COPY_NCDIM(var%dim(nrvdim), grid%hybi%dim(1))
       ldvar(ovec(3)) = var%dim(nrvdim)%len
       dpos(3) = 3
    END IF
    IF (ASSOCIATED(grid%timem%dim)) THEN
       nrvdim = nrvdim + 1
       CALL COPY_NCDIM(var%dim(nrvdim), grid%timem%dim(1))
    END IF

    var%ustep = grid%t
       

    ! PASS INFO AND DATA TO NARRAY
    IF (var%ndims > 0) THEN
       ALLOCATE(zdimlen(var%ndims), STAT=status)
       CALL ERRMSG(substr,status,13)
    ELSE
       ALLOCATE(zdimlen(1), STAT=status)
       CALL ERRMSG(substr,status,14)
       zdimlen(1) = 1
    END IF
    DO i=1, var%ndims
       zdimlen(i) = var%dim(i)%len  ! = 1 for UNLIM-DIM-ID
    END DO
    j = SIZE(zdimlen)
    CALL INIT_NARRAY(var%dat, j, zdimlen, VTYPE_DOUBLE)
    DEALLOCATE(zdimlen) ! um_ak_20140312

    IF (ASSOCIATED(var%dat%dim)) THEN
       npdlv = PRODUCT(var%dat%dim)
    ELSE
       RETURN
    END IF

    ! COPY DATA (LOOP OVER ELEMENTS)
    DO i=1, npdlv
       CALL ELEMENT(var%dat%dim, i, vec)
       nvec((ovec(4))) = PRODUCT(vec)
       IF (dpos(1) /= 0) THEN
          nvec((ovec(1))) = vec(dpos(1))
          nvec(ovec(4)) = nvec(ovec(4))/nvec(ovec(1))
       ELSE
          nvec(ovec(1)) = 1
       END IF
       IF (dpos(2) /= 0) THEN
          nvec((ovec(2))) = vec(dpos(2))
          nvec(ovec(4)) = nvec(ovec(4))/nvec(ovec(2))
       ELSE
          nvec(ovec(2)) = 1
       END IF
       IF (dpos(3) /= 0) THEN
          nvec((ovec(3))) = vec(dpos(3))
          nvec(ovec(4)) = nvec(ovec(4))/nvec(ovec(3))
       ELSE
          nvec(ovec(3)) = 1
       END IF
       var%dat%vd(i) = &
            dat(nvec(1), nvec(2), nvec(3), nvec(4)) 
       !
       DEALLOCATE(vec, STAT=status)
       NULLIFY(vec) 
       CALL ERRMSG(substr,status,2)
    END DO

  END SUBROUTINE RGTOOL_CONVERT_DAT2VAR
! ----------------------------------------------------------------------

! ----------------------------------------------------------------------
  SUBROUTINE RGTOOL_G2C(grid, hyam, hybm, p0, ps         &
                        ,hyai, hybi                      &
                        ,latm, lonm                      &
                        ,lati, loni, nlat, nlon, nlev    &
                        ,order)

    ! CONVERTS GRID INFORMATION (grid) TO ARRAYS
    ! hyam, hybm, hyai, hybi, latm, lati, lonm, loni, p0 -> 1D
    ! ps                                                 -> 2D
    !
    ! nlat, nlon, nlev are OPTIONAL parameters in order to specify
    ! the array-length (box-mid), in case the grid does not contain
    ! the respective dimension
    !
    ! Author: Patrick Joeckel, MPICH, October 2002

    USE MESSY_MAIN_GRID,         ONLY: t_geohybgrid
    USE MESSY_MAIN_GRID_NETCDF,  ONLY: ERRMSG

    IMPLICIT NONE

    INTRINSIC :: ASSOCIATED, PRESENT, SIZE

    ! I/O
    TYPE (t_geohybgrid),      INTENT(IN)        :: grid
    REAL(dp), DIMENSION(:),   POINTER, OPTIONAL :: hyam   ! hybrid-A-coeff.
    REAL(dp), DIMENSION(:),   POINTER, OPTIONAL :: hybm   ! hybrid-B-coeff.
    REAL(dp), DIMENSION(:),   POINTER, OPTIONAL :: p0     ! reference pressure
    REAL(dp), DIMENSION(:,:), POINTER, OPTIONAL :: ps     ! surface pressure
    REAL(dp), DIMENSION(:),   POINTER, OPTIONAL :: hyai   ! hybrid-A-coeff.
    REAL(dp), DIMENSION(:),   POINTER, OPTIONAL :: hybi   ! hybrid-B-coeff.
    REAL(dp), DIMENSION(:),   POINTER, OPTIONAL :: latm   ! latitude
    REAL(dp), DIMENSION(:),   POINTER, OPTIONAL :: lonm   ! longitude
    REAL(dp), DIMENSION(:),   POINTER, OPTIONAL :: lati   ! latitude
    REAL(dp), DIMENSION(:),   POINTER, OPTIONAL :: loni   ! longitude
    INTEGER,               INTENT(IN), OPTIONAL :: nlat
    INTEGER,               INTENT(IN), OPTIONAL :: nlon
    INTEGER,               INTENT(IN), OPTIONAL :: nlev
    CHARACTER(LEN=4),      INTENT(IN), OPTIONAL :: order  ! order of dimensions

    ! LOCAL
    CHARACTER(LEN=*), PARAMETER :: substr = 'RGTOOL_G2C'
    CHARACTER(LEN=4)            :: loc_order
    REAL(dp), DIMENSION(:,:,:,:), POINTER :: hha, hp0, hhb, hps, hlat, hlon
    INTEGER                     :: status

    NULLIFY(hha)
    NULLIFY(hp0)
    NULLIFY(hhb)
    NULLIFY(hps)
    NULLIFY(hlat)
    NULLIFY(hlon)
    
    IF (PRESENT(order)) THEN
       loc_order = order
    ELSE
       loc_order = 'xyzn'
    END IF

    IF (PRESENT(hyam)) THEN
       CALL RGTOOL_CONVERT(grid%hyam,hha,grid, order=loc_order)
       IF (ASSOCIATED(hha)) THEN
          ALLOCATE(hyam(SIZE(hha,2)), STAT=status)
          CALL ERRMSG(substr,status,1)
          hyam(:) = hha(1,:,1,1)
          DEALLOCATE(hha, STAT=status)
          NULLIFY(hha) 
          CALL ERRMSG(substr,status,2)
       ELSE
          IF (PRESENT(nlev)) THEN
             ALLOCATE(hyam(nlev), STAT=status)
             CALL ERRMSG(substr,status,3)
             hyam(:) = 0.0
          ELSE
             NULLIFY(hyam)
          END IF
       END IF
    END IF

    IF (PRESENT(hyai)) THEN
       CALL RGTOOL_CONVERT(grid%hyai,hha,grid, order=loc_order)
       IF (ASSOCIATED(hha)) THEN
          ALLOCATE(hyai(SIZE(hha,2)), STAT=status)
          CALL ERRMSG(substr,status,1)
          hyai(:) = hha(1,:,1,1)
          DEALLOCATE(hha, STAT=status)
          NULLIFY(hha) 
          CALL ERRMSG(substr,status,2)
       ELSE
          IF (PRESENT(nlev)) THEN
             ALLOCATE(hyai(nlev+1), STAT=status)
             CALL ERRMSG(substr,status,3)
             hyai(:) = 0.0
          ELSE
             NULLIFY(hyai)
          END IF
       END IF
    END IF

    IF (PRESENT(p0)) THEN
       CALL RGTOOL_CONVERT(grid%p0, hp0, grid, order=loc_order)
       IF (ASSOCIATED(hp0)) THEN
          ALLOCATE(p0(1), STAT=status)
          CALL ERRMSG(substr,status,4)
          p0 = hp0(1,1,1,1)
          DEALLOCATE(hp0, STAT=status)
          NULLIFY(hp0) 
          CALL ERRMSG(substr,status,5)
       ELSE
          ALLOCATE(p0(1), STAT=status)
          CALL ERRMSG(substr,status,6)
          p0(1) = 0.0
       ENDIF
    END IF

    IF (PRESENT(hybm)) THEN
       CALL RGTOOL_CONVERT(grid%hybm, hhb, grid, order=loc_order)
       IF (ASSOCIATED(hhb)) THEN
          ALLOCATE(hybm(SIZE(hhb,2)), STAT=status)
          CALL ERRMSG(substr,status,7)
          hybm(:) = hhb(1,:,1,1)
          DEALLOCATE(hhb, STAT=status)
          NULLIFY(hhb) 
          CALL ERRMSG(substr,status,8)
       ELSE
          IF (PRESENT(nlev)) THEN
             ALLOCATE(hybm(nlev), STAT=status)
             CALL ERRMSG(substr,status,9)
             hybm(:) = 0.0
          ELSE
             NULLIFY(hybm)
          END IF
       END IF
    END IF

    IF (PRESENT(hybi)) THEN
       CALL RGTOOL_CONVERT(grid%hybi, hhb, grid, order=loc_order)
       IF (ASSOCIATED(hhb)) THEN
          ALLOCATE(hybi(SIZE(hhb,2)), STAT=status)
          CALL ERRMSG(substr,status,7)
          hybi(:) = hhb(1,:,1,1)
          DEALLOCATE(hhb, STAT=status)
          NULLIFY(hhb) 
          CALL ERRMSG(substr,status,8)
       ELSE
          IF (PRESENT(nlev)) THEN
             ALLOCATE(hybi(nlev+1), STAT=status)
             CALL ERRMSG(substr,status,9)
             hybi(:) = 0.0
          ELSE
             NULLIFY(hybi)
          END IF
       END IF
    END IF

    IF (PRESENT(ps)) THEN
       CALL RGTOOL_CONVERT(grid%ps, hps, grid, order=loc_order)
       IF (ASSOCIATED(hps)) THEN
          ALLOCATE(ps(SIZE(hps,1),SIZE(hps,4)), STAT=status)
          CALL ERRMSG(substr,status,10)
          ps(:,:) = hps(:,1,1,:)
          DEALLOCATE(hps, STAT=status)
          NULLIFY(hps) 
          CALL ERRMSG(substr,status,11)
       ELSE
          IF (PRESENT(nlat).AND.PRESENT(nlon)) THEN
             ALLOCATE(ps(nlon,nlat), STAT=status)
             CALL ERRMSG(substr,status,12)
             ps(:,:) = 0.0
          ELSE
             NULLIFY(ps)
          END IF
       END IF
    END IF

    IF (PRESENT(latm)) THEN
       CALL RGTOOL_CONVERT(grid%latm, hlat, grid, order=loc_order)
       IF (ASSOCIATED(hlat)) THEN
          ALLOCATE(latm(SIZE(hlat,4)), STAT=status)
          CALL ERRMSG(substr,status,13)
          latm(:) = hlat(1,1,1,:)
          DEALLOCATE(hlat, STAT=status)
          NULLIFY(hlat) 
          CALL ERRMSG(substr,status,14)
       ELSE
          IF (PRESENT(nlat)) THEN
             ALLOCATE(latm(nlat), STAT=status)
             CALL ERRMSG(substr,status,15)
             latm(:) = 0.0
          ELSE
             NULLIFY(latm)
          END IF
       END IF
    END IF

    IF (PRESENT(lati)) THEN
       CALL RGTOOL_CONVERT(grid%lati, hlat, grid, order=loc_order)
       IF (ASSOCIATED(hlat)) THEN
          ALLOCATE(lati(SIZE(hlat,4)), STAT=status)
          CALL ERRMSG(substr,status,13)
          lati(:) = hlat(1,1,1,:)
          DEALLOCATE(hlat, STAT=status)
          NULLIFY(hlat) 
          CALL ERRMSG(substr,status,14)
       ELSE
          IF (PRESENT(nlat)) THEN
             ALLOCATE(lati(nlat+1), STAT=status)
             CALL ERRMSG(substr,status,15)
             lati(:) = 0.0
          ELSE
             NULLIFY(lati)
          END IF
       END IF
    END IF

    IF (PRESENT(lonm)) THEN
       CALL RGTOOL_CONVERT(grid%lonm, hlon, grid, order=loc_order)
       IF (ASSOCIATED(hlon)) THEN
          ALLOCATE(lonm(SIZE(hlon,1)), STAT=status)
          CALL ERRMSG(substr,status,16)
          lonm(:) = hlon(:,1,1,1)
          DEALLOCATE(hlon, STAT=status)
          NULLIFY(hlon) 
          CALL ERRMSG(substr,status,17)
       ELSE
          IF (PRESENT(nlon)) THEN
             ALLOCATE(lonm(nlon), STAT=status)
             CALL ERRMSG(substr,status,18)
             lonm(:) = 0.0
          ELSE
             NULLIFY(lonm)
          END IF
       END IF
    END IF

    IF (PRESENT(loni)) THEN
       CALL RGTOOL_CONVERT(grid%loni, hlon, grid, order=loc_order)
       IF (ASSOCIATED(hlon)) THEN
          ALLOCATE(loni(SIZE(hlon,1)), STAT=status)
          CALL ERRMSG(substr,status,16)
          loni(:) = hlon(:,1,1,1)
          DEALLOCATE(hlon, STAT=status)
          NULLIFY(hlon) 
          CALL ERRMSG(substr,status,17)
       ELSE
          IF (PRESENT(nlon)) THEN
             ALLOCATE(loni(nlon+1), STAT=status)
             CALL ERRMSG(substr,status,18)
             loni(:) = 0.0
          ELSE
             NULLIFY(loni)
          END IF
       END IF
    END IF

  END SUBROUTINE RGTOOL_G2C
! ----------------------------------------------------------------------

! um_ak_20150506+
! MOVED HERE FROM GRID_TRAFO_SCRP
  ! ===========================================================================
!  SUBROUTINE aline_array(status, var2d, grid, var1d)
  SUBROUTINE ALINE_ARRAY(status, var2d, var1d)

    IMPLICIT NONE

    ! I/O
    INTEGER,                  INTENT(OUT) :: status
    REAL(dp), DIMENSION(:,:), INTENT(IN)  :: var2d
!    TYPE(t_scrip_grid)      , INTENT(IN)  :: grid
    REAL(dp), DIMENSION(:),   POINTER     :: var1d
    ! LOCAL
    INTEGER :: i,j,n
    INTEGER :: idim
    INTEGER :: isize

    status = 3999

    isize= SIZE(var2d,1) * SIZE(var2d,1)
     IF (ASSOCIATED(var1d)) DEALLOCATE(var1d)
    ALLOCATE(var1d(isize))
!    ALLOCATE(var1d(grid%size))
    idim = SIZE(var2d,1)
    DO i=1, idim
       DO j=1, SIZE(var2d,2)
          n = (j-1) * idim + i
          var1d(n) = var2d(i,j)
       END DO
    END DO

    status = 0

  END SUBROUTINE ALINE_ARRAY
  ! ========================================================================

  ! ===========================================================================
  SUBROUTINE DEALINE_ARRAY(status, dim1, dim2, var1d, var2d)

    IMPLICIT NONE

    ! I/O
    INTEGER,                 INTENT(OUT) :: status
    INTEGER,                 INTENT(IN)  :: dim1, dim2
    REAL(dp), DIMENSION(:),  INTENT(IN)  :: var1d
    REAL(dp), DIMENSION(:,:), POINTER    :: var2d
    ! LOCAL
    INTEGER :: i,j,n

    status = 3999

    IF (dim1 *dim2 /= SIZE(var1d)) THEN
       status = 3022 ! sizes incompatible
       RETURN
    ENDIF

    IF (ASSOCIATED(var2d)) DEALLOCATE(var2d)
    ALLOCATE(var2d(dim1,dim2))
    DO i=1, dim1
       DO j=1, dim2
          n = (j-1) * dim1 + i
          var2d(i,j) = var1d(n)
       END DO
    END DO

    status = 0
    
  END SUBROUTINE DEALINE_ARRAY
  ! ========================================================================
! um_ak_20150506-
! ----------------------------------------------------------------------
! um_ak_20141007+
! moved here from mmd2way_child

  SUBROUTINE SET_SURFACE_PRESSURE(status, grid, press, lcalc)

    USE MESSY_MAIN_GRID,         ONLY: t_geohybgrid
    USE MESSY_MAIN_GRID_NETCDF,  ONLY: NULL_VARID, NF90_FLOAT, VTYPE_REAL &
                                     , POSITION, QDEF_NCVAR, ADD_NCATT &
                                     , ERRMSG, INIT_NARRAY, INIT_NCVAR &
                                     , COPY_NCDIM
    IMPLICIT NONE

    ! I/O
    INTEGER,                  INTENT(OUT)          :: status
    TYPE(t_geohybgrid),       INTENT(INOUT)        :: grid
    REAL(dp), DIMENSION(:,:), INTENT(IN)           :: press
    ! limit variable pressure to subparts
    LOGICAL,  DIMENSION(:,:), INTENT(IN), OPTIONAL :: lcalc

    ! LOCAL
    CHARACTER(LEN=*), PARAMETER :: substr = 'set_surface_pressure'
    INTEGER :: i, j, id, jd

    CALL INIT_NCVAR(grid%ps)
    grid%ps%name  = 'ps'
    grid%ps%id    = NULL_VARID
    grid%ps%xtype = NF90_FLOAT
    ! ... dimensions
    IF (.NOT. ASSOCIATED(grid%ps%dim)) THEN
       grid%ps%ndims = 2
       ALLOCATE(grid%ps%dim(grid%ps%ndims), STAT=status)
       CALL ERRMSG('INTERFACE_GEOHYBGRID',status,5)
    ELSE
       IF (SIZE(grid%ps%dim) /= 2) THEN
          status = 2100 
          RETURN
          ! 'ps grid dimensions do not match'
       END IF
    END IF

    ! horizontal non-rotated grid has been defined 
    CALL COPY_NCDIM(grid%ps%dim(1), grid%lonm%dim(1))
    CALL COPY_NCDIM(grid%ps%dim(2), grid%latm%dim(1))

    ! ... data
    CALL INIT_NARRAY(grid%ps%dat, grid%ps%ndims       &
         , (/grid%ps%dim(1)%len, grid%ps%dim(2)%len/) &
         , VTYPE_REAL)
    DO i=1,grid%ps%dim(1)%len
       DO j=1, grid%ps%dim(2)%len
          IF (PRESENT(lcalc)) THEN
             IF (lcalc(i,j)) THEN
                grid%ps%dat%vr(&
                     POSITION((/grid%ps%dim(1)%len, grid%ps%dim(2)%len/)&
                     ,(/i,j/)))  = press(i,j)
             ELSE
                grid%ps%dat%vr(&
                     POSITION((/grid%ps%dim(1)%len, grid%ps%dim(2)%len/) &
                     ,(/i,j/)))   =  101325.0_dp
             END IF
          ELSE
             grid%ps%dat%vr(&
                  POSITION((/grid%ps%dim(1)%len, grid%ps%dim(2)%len/)&
                  ,(/i,j/)))  = press(i,j)
          END IF
       END DO
    END DO

    status = 0

  END SUBROUTINE SET_SURFACE_PRESSURE
! ---------------------------------------------------------------------------
! um_ak_20141007-
END MODULE MESSY_MAIN_GRID_TOOLS
